@@ -2,7 +2,7 @@ import inspect
 import dis
 import ctypes
 from collections import namedtuple
-from functools import partial
+import functools
 from itertools import groupby
 from types import CodeType, FunctionType
 import logging
@@ -22,13 +22,19 @@ from .py import (
     ptr_frame_stack_top,
     ptr_frame_block_stack_bottom,
     ptr_frame_block_stack_top,
-    put_NULL,
-    disassemble,
 )
 from .minias import _dis, long2bytes
+from .morph import morph_stack
+from .primitives import NULL
+
 
 locals().update(dis.opmap)
-EXCEPT_HANDLER = 257
+
+
+class partial(functools.partial):
+    def __str__(self):
+        return f"partial({self.func}, positional={len(self.args)}, kw=[{', '.join(self.keywords)}])"
+    __repr__ = __str__
 
 
 def _overlapping(s1, l1, s2, l2):
@@ -54,7 +60,7 @@ class CodePatcher(dict):
     def commit(self):
         logging.debug(f"Commit patch to <{self._code.co_name}>")
         for i in self._diff():
-            logging.debug(''.join(i))
+            logging.log(5, ''.join(i))
         code = self._code.co_code
         code_view = Mem.view(code)
         for pos, patch in self.items():
@@ -132,16 +138,6 @@ def expand_long(c):
     return bytes(result)
 
 
-class NULL:
-    """Represents NULL"""
-    def __new__(cls):
-        return cls
-
-    def __str__(self):
-        return "<NULL>"
-    __repr__ = __str__
-
-
 def get_value_stack_from_beacon(frame, beacon, expand=0, null=NULL()):
     """
     Collects frame stack using beacon as
@@ -233,7 +229,7 @@ def get_block_stack(frame):
     return result
 
 
-class FrameSnapshot(namedtuple("FrameSnapshot", ("code", "pos", "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack"))):
+class FrameSnapshot(namedtuple("FrameSnapshot", ("scope", "code", "pos", "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack"))):
     """A snapshot of python frame"""
     slots = ()
     def __repr__(self):
@@ -245,7 +241,7 @@ class FrameSnapshot(namedtuple("FrameSnapshot", ("code", "pos", "v_stack", "v_lo
                 contents.append(f"{i}: not set")
             else:
                 contents.append(f"{i}: {len(v):d}")
-        return f'FrameSnapshot {code.co_name} at "{code.co_filename}"+{code.co_firstlineno} @{self.pos:d} {" ".join(contents)}'
+        return f'FrameSnapshot {self.scope} -> {code.co_name} "{code.co_filename}"+{code.co_firstlineno} @{self.pos:d} {" ".join(contents)}'
 
 
 def p_jump_to(pos, patcher, f_next, jx=JX):
@@ -327,8 +323,8 @@ def p_place_beacon(beacon, patcher, f_next):
     logging.debug(f"place_beacon {beacon}: patching ...")
     patcher.patch_current([
         UNPACK_SEQUENCE, 2,
-        CALL_FUNCTION, 0,  # calls _payload1
-        CALL_FUNCTION, 0,  # calls whatever follows
+        CALL_FUNCTION, 0,  # calls f_next
+        CALL_FUNCTION, 0,  # calls what f_next returns
     ], 2)
     patcher.commit()
     logging.debug(f"place_beacon {beacon}: ⏎ ({f_next}, {beacon})")
@@ -427,11 +423,7 @@ def snapshot(frame, finalize, method="inject"):
     while frame is not None:  # iterate over frame stack
         logging.info(f"Frame: {frame}")
 
-        # check globals and builtins
-        if prev_globals is None:
-            prev_globals = frame.f_globals
-        else:
-            assert prev_globals is frame.f_globals
+        # check builtins
         if prev_builtins is None:
             prev_builtins = frame.f_builtins
         else:
@@ -440,14 +432,16 @@ def snapshot(frame, finalize, method="inject"):
         # save locals, globals, etc.
         logging.info("  saving snapshot ...")
         fs = FrameSnapshot(
+            scope=inspect.getmodule(frame),
             code=frame.f_code,
             pos=frame.f_lasti,
             v_stack=None if method == "inject" else get_value_stack(frame),
             v_locals=frame.f_locals.copy(),
-            v_globals=prev_globals,
-            v_builtins=prev_builtins,
+            v_globals=frame.f_globals,
+            v_builtins=frame.f_builtins,
             block_stack=get_block_stack(frame),
         )
+        logging.info(f"    scope: {fs.scope}")
         logging.info(f"    code: {fs.code}")
         logging.info(f"    pos: {fs.pos}")
         logging.info(f"    stack: {len(fs.v_stack) if fs.v_stack is not None else 'none'}")
@@ -460,7 +454,6 @@ def snapshot(frame, finalize, method="inject"):
                 logging.info(f"      {i}")
         else:
             logging.info("      (empty)")
-
         result.append(fs)
 
         if method == "inject":  # prepare patchers
@@ -473,6 +466,7 @@ def snapshot(frame, finalize, method="inject"):
             p_jump_to(0, patcher, None)  # make room for patches immediately
             chain.append(partial(p_place_beacon, beacon, patcher))  # place the beacon
             chain.append(partial(notify, frame))  # collect value stack
+            chain.append(partial(p_jump_to, 0, patcher))  # jump once more to make more room
             chain.append(partial(p_exit_block_stack, fs.block_stack, patcher))  # exit from "finally" statements
             chain.append(partial(p_jump_to, rtn_pos - 2, patcher))  # jump 1 opcode before return
             chain.append(partial(
@@ -498,21 +492,23 @@ def snapshot(frame, finalize, method="inject"):
         return result
 
 
-def unpickle_generator(code):
+def unpickle_generator(code, scope):
     """
     Unpickles the generator.
 
     Parameters
     ----------
-    code : Codetype
-        The morph code.
+    code : CodeType
+        Generator (morph) code.
+    scope
+        Generator scope.
 
     Returns
     -------
     result
         The generator.
     """
-    return FunctionType(code, globals())()
+    return FunctionType(code, scope.__dict__)()
 
 
 def _():
@@ -531,174 +527,8 @@ def pickle_generator(pickler, obj):
     obj
         The generator.
     """
-    code = morph_stack(snapshot(obj.gi_frame, None, method="direct"), globals=False, flags=0x20)
-    pickler.save_reduce(
-        unpickle_generator,
-        (code,),
-        obj=obj,
-    )
-
-
-def morph_execpoint(p, nxt, pack=None, unpack=None, globals=False, fake_return=True,
-                    flags=0):
-    """
-    Prepares a code object which morphs into the desired state
-    and continues the execution afterwards.
-
-    Parameters
-    ----------
-    p : execpoint
-        The execution point to morph into.
-    nxt : CodeType
-        The code object which develops the stack further.
-    pack : Callable, None
-        A method turning objects into bytes (serializer)
-        locally.
-    unpack : tuple, None
-        A 2-tuple `(module_name, method_name)` specifying
-        the method that morph uses to unpack the data.
-    globals : bool
-        If True, unpacks globals.
-    fake_return : bool
-        If set, fakes returning None by putting None on top
-        of the stack. This will be ignored if nxt is not
-        None.
-    flags : int
-        Code object flags.
-
-    Returns
-    -------
-    result : CodeType
-        The resulting morph.
-    """
-    assert pack is None and unpack is None or pack is not None and unpack is not None,\
-        "Either both or none pack and unpack arguments have be specified"
-    logging.info(f"Preparing a morph into execpoint {p} pack={pack is not None} ...")
-    code = disassemble(p.code)
-    code.pos = 0
-    code.c("Header")
-    code.nop(b'mrph')  # signature
-    f_code = p.code
-    new_stacksize = f_code.co_stacksize
-
-    if pack:
-        unpack_mod, unpack_method = unpack
-        code.c(f"from {unpack_mod} import {unpack_method}")
-        unpack = code.varnames('.:unpack:.')  # non-alphanumeric = unlikely to exist as a proper variable
-        code.I(LOAD_CONST, 0)
-        code.I(LOAD_CONST, (unpack_method,))
-        code.I(IMPORT_NAME, unpack_mod)
-        code.I(IMPORT_FROM, unpack_method)
-        code.i(STORE_FAST, unpack)
-
-        def _LOAD(_what):
-            code.i(LOAD_FAST, unpack)
-            code.I(LOAD_CONST, pack(_what))
-            code.i(CALL_FUNCTION, 1)
-    else:
-        def _LOAD(_what):
-            code.I(LOAD_CONST, _what)
-
-    scopes = [(p.v_locals, STORE_FAST, "locals")]
-    if globals:
-        scopes.append((p.v_globals, STORE_GLOBAL, "globals"))
-    for _dict, _STORE, log_name in scopes:
-        logging.info(f"  {log_name} ...")
-        if len(_dict) > 0:
-            code.c(f"{log_name} = ...")
-            klist, vlist = zip(*_dict.items())
-            _LOAD(vlist)
-            code.i(UNPACK_SEQUENCE, len(vlist))
-            for k in klist:
-                # k = v
-                code.I(_STORE, k)
-            new_stacksize = max(new_stacksize, len(vlist))
-
-    # stack
-    if len(p.v_stack) > 0:
-        for k, v_stack in groupby(p.v_stack, lambda x: x is not NULL):
-            if k:
-                code.c("*stack")
-                v_stack = tuple(v_stack)[::-1]
-                _LOAD(v_stack)
-                code.i(UNPACK_SEQUENCE, len(v_stack))
-            else:
-                code.c("*NULL")
-                for _ in v_stack:
-                    put_NULL(code)
-
-    # block stack
-    if len(p.block_stack) > 0:
-        code.c("block_stack")
-        for i, (type, handler, level) in enumerate(p.block_stack):
-            if type == SETUP_FINALLY:
-                code.i(SETUP_FINALLY, 0, jump_to=code.by_pos(handler * JX))
-            elif type == EXCEPT_HANDLER:
-                raise NotImplementedError(f"'except:' not implemented")
-            else:
-                raise NotImplementedError(f"Unknown block type={type} ({dis.opname.get(type, 'unknown opcode')})")
-
-    if nxt is not None:
-        # call nxt which is a code object
-        code.c(f"nxt()")
-
-        # load code object
-        _LOAD(nxt)
-        code.I(LOAD_CONST, None)  # function name
-        code.i(MAKE_FUNCTION, 0)  # turn code object into a function
-        code.i(CALL_FUNCTION, 0)  # call it
-    elif fake_return:
-        code.c(f"fake return None")
-        code.I(LOAD_CONST, None)  # fake nxt returning None
-
-    # now jump to the previously saved position
-    code.c(f"goto saved pos")
-    code.i(JUMP_ABSOLUTE, 0, jump_to=code.by_pos(p.pos + 2))
-
-    code.c(f"---------------------")
-    code.c(f"The original bytecode")
-    code.c(f"---------------------")
-    result = CodeType(
-        0,
-        0,
-        0,
-        len(code.varnames),
-        new_stacksize + 1,
-        flags,
-        code.get_bytecode(),
-        tuple(code.consts),
-        tuple(code.names),
-        tuple(code.varnames),
-        f_code.co_filename,  # TODO: something smarter should be here
-        f_code.co_name,
-        f_code.co_firstlineno,
-        f_code.co_lnotab,
-        )
-    logging.info(f"resulting morph:\n{str(code)}")
-    return result
-
-
-def morph_stack(frame_data, globals=True, **kwargs):
-    """
-    Morphs the stack.
-
-    Parameters
-    ----------
-    frame_data : list
-        States of all individual frames.
-    kwargs
-        Arguments to morph_execpoint.
-
-    Returns
-    -------
-    result : CodeType
-        The resulting morph for the root frame.
-    """
-    prev = None
-    for i, frame in enumerate(frame_data):
-        logging.info(f"Preparing morph #{i:d}")
-        prev = morph_execpoint(frame, prev, globals=globals and frame is frame_data[-1], **kwargs)
-    return prev
+    morph_data = morph_stack(snapshot(obj.gi_frame, None, method="direct"), root=False, flags=0x20)
+    pickler.save_reduce(unpickle_generator, morph_data, obj=obj)
 
 
 def dump(file, **kwargs):
@@ -713,7 +543,9 @@ def dump(file, **kwargs):
         Arguments to `dill.dump`.
     """
     def serializer(stack_data):
-        dill.dump(FunctionType(morph_stack(stack_data), globals()), file, **kwargs)
+        root_code, root_scope = morph_stack(stack_data)
+        # TODO: the scope probably needs to be fixed
+        dill.dump(FunctionType(root_code, {}), file, **kwargs)
     return snapshot(
         inspect.currentframe().f_back,
         finalize=serializer,
@@ -721,6 +553,18 @@ def dump(file, **kwargs):
 load = dill.load
 
 
+def is_python_interactive():
+    """
+    Determines if python is in interactive mode.
+
+    Returns
+    -------
+    is_interactive : bool
+        True if in interactive.
+    """
+    return "ps1" in dir(sys)
+
+
 def bash_inline_create_file(name, contents):
     """
     Turns a file into bash command.
@@ -740,10 +584,10 @@ def bash_inline_create_file(name, contents):
     return f"echo {quote(base64.b64encode(contents).decode())} | base64 -d > {quote(name)}"
 
 
-def shell_teleport(*shell_args, python=None, before="cd $(mktemp -d)",
+def tp_shell(*shell_args, python="python", before="cd $(mktemp -d)",
         pyc_fn="payload.pyc", shell_delimeter="; ", pack_file=bash_inline_create_file,
         pack_object=dill.dumps, unpack_object=("dill", "loads"),
-        _frame=None, **kwargs):
+        detect_interactive=True, files=None, _frame=None, **kwargs):
     """
     Teleport into another shell.
 
@@ -768,32 +612,38 @@ def shell_teleport(*shell_args, python=None, before="cd $(mktemp -d)",
     unpack_object : tuple, None
         A 2-tuple `(module_name, method_name)` specifying
         the method that morph uses to unpack the data.
+    detect_interactive : bool
+        If True, attempts to detect the interactive mode
+        and to open an interactive session remotely.
+    files : list
+        A list of files to teleport alongside.
     _frame
         The frame to collect.
     kwargs
         Other arguments to `subprocess.run`.
-
-    Returns
-    -------
-    None
     """
-    if python is None:
-        python = sys.executable
     payload = []
     if not isinstance(before, (list, tuple)):
         payload.append(before)
     else:
         payload.extend(before)
+    if files is None:
+        files = []
+
+    python_flags = []
+    if is_python_interactive():
+        python_flags.append("-i")
 
     def _teleport(stack_data):
         """Will be executed after the snapshot is done."""
+        nonlocal files
         logging.info("Snapshot done, composing morph ...")
-        code = morph_stack(stack_data, pack=pack_object, unpack=unpack_object)  # compose the code object
+        code, _ = morph_stack(stack_data, pack=pack_object, unpack=unpack_object)  # compose the code object
         logging.info("Creating pyc ...")
-        files = {pyc_fn: _code_to_timestamp_pyc(code)}  # turn it into pyc
+        files = {pyc_fn: _code_to_timestamp_pyc(code), **{k: open(k, 'rb').read() for k in files}}  # turn it into pyc
         for k, v in files.items():
             payload.append(pack_file(k, v))  # turn files into shell commands
-        payload.append(f"{python} {pyc_fn}")  # execute python
+        payload.append(f"{python} {' '.join(python_flags)} {pyc_fn}")  # execute python
 
         # pipe the output and exit
         logging.info("Executing the payload ...")
@@ -805,11 +655,12 @@ def shell_teleport(*shell_args, python=None, before="cd $(mktemp -d)",
         inspect.currentframe().f_back if _frame is None else _frame,
         finalize=_teleport,
     )
-bash_teleport = shell_teleport
+tp_bash = tp_shell
 
 
-def dummy_teleport(**kwargs):
+def tp_dummy(**kwargs):
     """A dummy teleport into another python process in current environment."""
-    return bash_teleport("bash", "-c", _frame=inspect.currentframe().f_back, **kwargs)
-
+    if "python" not in kwargs:
+        kwargs["python"] = sys.executable
+    return tp_shell("bash", "-c", _frame=inspect.currentframe().f_back, **kwargs)
 
