@@ -28,7 +28,6 @@ class Instruction:
     arg: int
     pos: int = 0
     len: int = 2
-    jump_target: list = field(default_factory=list)
     jump_to: object = None
 
     @property
@@ -39,29 +38,60 @@ class Instruction:
     def is_jump(self):
         return self.opcode in dis.hasjabs
 
+    @property
+    def is_any_jump(self):
+        return self.is_jrel or self.is_jump
+
+    @property
+    def pos_last(self):
+        return self.pos + self.len - 2
+
     @property
     def bytes(self):
-        arg_bytes = long2bytes(self.arg)
-        assert len(arg_bytes) * 2 == self.len, f"len({arg_bytes}) != {self.len}"
+        arg_bytes = long2bytes(max(self.arg, 0))
         result = []
         for i in arg_bytes[:-1]:
             result.extend((EXTENDED_ARG, i))
         result.extend((self.opcode, arg_bytes[-1]))
         return bytes(result)
 
-    def __repr__(self):
+    def compute_jump(self, jx):
+        if self.is_jrel:
+            return self.arg * jx + self.pos_last + 2
+        elif self.is_jump:
+            return self.arg * jx
+        else:
+            return None
+
+    def assert_valid(self, prev, lookup=None, jx=None):
+        assert self.arg >= 0, f"arg is negative: {self.arg}"
+        assert len(self.bytes) == self.len, f"len is invalid: len({repr(self.bytes)}) != {self.len}"
+        if prev is not None:
+            assert prev.pos + prev.len == self.pos, f"pos is invalid: {prev.pos}(prev.pos) + {prev.len}(prev.len) != {self.pos}(self.pos)"
+        else:
+            assert self.pos == 0, f"pos is non-zero: {self.pos}"
+        if lookup is not None and self.is_any_jump:
+            jump_points_to = lookup.get(self.compute_jump(jx), None)
+            assert jump_points_to is self.jump_to, f"jump_to is invalid: {repr(self.jump_to)} vs {repr(jump_points_to)}"
+
+    def __str__(self):
         return f"{self.pos:>6d} {_trunc(dis.opname[self.opcode], 18):<18} {self.arg:<16d}"
 
+    def __repr__(self):
+        return f"{dis.opname[self.opcode]}({self.arg}, pos={self.pos}, len={self.len})"
 
 @dataclass
 class Comment:
     """Represents a comment"""
     text: str
-    def __post_init__(self):
-        assert len(self.text) < 35
+    @property
+    def printable_text(self):
+        if len(self.text) < 36:
+            return self.text
+        return f"{self.text[:32]}..."
 
     def __repr__(self):
-        return f"       {self.text}" + " " * (35 - len(self.text))
+        return f"       {self.printable_text}" + " " * (35 - len(self.text))
 
 
 class CList(list):
@@ -75,12 +105,12 @@ class CList(list):
 
 
 class Bytecode(list):
-    def __init__(self, opcodes, names, varnames, consts, jx=1):
+    def __init__(self, opcodes, co_names, co_varnames, co_consts, jx=1):
         super().__init__(opcodes)
         self.pos = len(self)
-        self.names = CList(names)
-        self.varnames = CList(varnames)
-        self.consts = CList(consts)
+        self.co_names = CList(co_names)
+        self.co_varnames = CList(co_varnames)
+        self.co_consts = CList(co_consts)
         self._jx = jx
 
     @classmethod
@@ -88,6 +118,12 @@ class Bytecode(list):
         if isinstance(arg, FunctionType):
             arg = arg.__code__
         code = arg.co_code
+        marks = dis.findlinestarts(arg)
+        try:
+            lines = open(arg.co_filename, 'r').readlines()[arg.co_firstlineno - 1:]
+            marks = list((i_opcode, i_line, lines[i_line - 1].strip()) for (i_opcode, i_line) in marks)
+        except (TypeError, OSError, IndexError):
+            marks = None
         result = cls([], arg.co_names, arg.co_varnames, arg.co_consts, **kwargs)
         arg = 0
         _len = 0
@@ -95,7 +131,19 @@ class Bytecode(list):
             arg = arg * 0x100 + _arg
             _len += 2
             if opcode != EXTENDED_ARG:
-                result.i(opcode, arg, pos * 2 - _len + 2, _len)
+                start_pos = pos * 2 - _len + 2
+
+                if marks is not None:
+                    for i_mark, (mark_opcode, mark_lineno, mark_text) in enumerate(marks):
+                        if mark_opcode <= start_pos:
+                            result.c(f"{mark_lineno:<3d} {mark_text}")
+                        else:
+                            marks = marks[i_mark:]
+                            break
+                    else:
+                        marks = []
+
+                result.i(opcode, arg, start_pos, _len)
                 arg = _len = 0
         result.eval_jumps()
         return result
@@ -117,13 +165,17 @@ class Bytecode(list):
 
     def I(self, opcode, arg, *args, **kwargs):
         if opcode in dis.hasconst:
-            return self.i(opcode, self.consts(arg), *args, **kwargs)
+            return self.i(opcode, self.co_consts(arg), *args, **kwargs)
         elif opcode in dis.hasname:
-            return self.i(opcode, self.names(arg), *args, **kwargs)
+            return self.i(opcode, self.co_names(arg), *args, **kwargs)
         elif opcode in dis.haslocal:
-            return self.i(opcode, self.varnames(arg), *args, **kwargs)
+            return self.i(opcode, self.co_varnames(arg), *args, **kwargs)
+        elif opcode in dis.hasjrel + dis.hasjabs:
+            result = self.i(opcode, None)
+            result.jump_to = arg
+            return result
         else:
-            raise ValueError(f"Unknown opcode: {dis.opnames[opcode]}")
+            raise ValueError(f"Unknown opcode: {dis.opname[opcode]}")
 
     def nop(self, arg):
         arg = bytes(arg)
@@ -144,46 +196,47 @@ class Bytecode(list):
 
     def eval_jumps(self):
         lookup = {i.pos: i for i in self.iter_opcodes()}
-        for i in self:
-            if i.is_jrel:
-                target = lookup[i.arg * self._jx + i.pos + 2]
-            elif i.is_jump:
-                target = lookup[i.arg * self._jx]
-            else:
-                target = None
-            if target is not None:
-                target.jump_target.append(i)
-                i.jump_to = target
+        for i in self.iter_opcodes():
+            if i.is_any_jump and i.arg is not None:
+                i.jump_to = lookup[i.compute_jump(self._jx)]
 
     def assign_pos(self):
         pos = 0
-        result = False
         for i in self.iter_opcodes():
-            if i.pos != pos:
-                result = True
             i.pos = pos
             pos += i.len
-        return result
 
     def assign_jump_args(self):
         for i in self.iter_opcodes():
-            if i.is_jump:
-                i.arg = i.jump_to.pos // self._jx
-            elif i.is_jrel:
-                i.arg = (i.jump_to.pos - i.pos - 2) // self._jx
+            if i.is_any_jump and i.jump_to is not None:
+                if i.is_jump:
+                    i.arg = i.jump_to.pos // self._jx
+                elif i.is_jrel:
+                    i.arg = (i.jump_to.pos - i.pos_last - 2) // self._jx
 
     def assign_len(self):
         for i in self.iter_opcodes():
-            i.len = 2 * len(long2bytes(i.arg))
+            i.len = len(i.bytes)
+
+    def assert_valid(self):
+        prev = None
+        lookup = {i.pos: i for i in self.iter_opcodes()}
+        for opcode in self.iter_opcodes():
+            opcode.assert_valid(prev, lookup=lookup, jx=self._jx)
+            prev = opcode
 
     def get_bytecode(self):
-        for i in range(4):
+        for i in range(5):
             self.assign_jump_args()
             self.assign_len()
-            if self.assign_pos() is False:
+            self.assign_pos()
+            try:
+                self.assert_valid()
                 break
+            except AssertionError:
+                pass
         else:
-            raise ValueError("Failed to re-assemble")
+            self.assert_valid()  # re-raise
         return b''.join(i.bytes for i in self.iter_opcodes())
 
     def __str__(self):
@@ -221,19 +274,33 @@ class Bytecode(list):
                     else:
                         _str.append(" ")
                 lines.append(str(i) + ''.join(_str))
+
+            if isinstance(i, Instruction):
+                represented, arg = _repr_arg(i.opcode, i.arg, self)
+                if represented:
+                    arg_repr = repr(arg)
+                    if len(arg_repr) > 24:
+                        arg_repr = f"<{type(arg).__name__} instance>"
+                    lines[-1] = lines[-1] + f" ({arg_repr})"
         return '\n'.join(lines)
 
 
+def _repr_arg(opcode, arg, code):
+    if opcode in dis.hasconst:
+        return True, code.co_consts[arg]
+    elif opcode in dis.haslocal:
+        return True, code.co_varnames[arg]
+    elif opcode in dis.hasname:
+        return True, code.co_names[arg]
+    else:
+        return False, arg
+
+
 def _repr_opcode(opcode, arg, code):
     head = f"{dis.opname[opcode]:>20} {arg: 3d}"
-    if opcode == LOAD_CONST:
-        return f"{head} {'(' + repr(code.co_consts[arg]) + ')':<12}"
-    elif opcode in (LOAD_FAST, STORE_FAST):
-        return f"{head} {code.co_varnames[arg]:<12}"
-    elif opcode in (LOAD_NAME, STORE_NAME):
-        return f"{head} {code.co_names[arg]:<12}"
-    elif opcode in (LOAD_GLOBAL, STORE_GLOBAL):
-        return f"{head} {'(' + repr(code.co_names[arg]) + ')':<12}"
+    represented, val = _repr_arg(opcode, arg, code)
+    if represented:
+        return f"{head} {'(' + repr(val) + ')':<12}"
     else:
         return f"{head}" + " " * 13
 
